from enum import Enum

from .bioformats import BioFormatsReader
from .openslide import OpenSlideReader
from .cucim import CuCIMReader
from .tiffslide import TiffSlideReader


class Reader(Enum):
    OPENSLIDE = "openslide"
    CUCIM = "cucim"
    TIFFSLIDE = "tiffslide"
    BIOFORMATS = "bioformats"


InstallationTips = {
    Reader.OPENSLIDE: "Please install OpenSlide with conda/mamba, `conda install openslide-python`",
    Reader.CUCIM: "Please see `https://github.com/rapidsai/cucim` for installation instructions",
    Reader.TIFFSLIDE: "`pip install tiffslide`",
    Reader.BIOFORMATS: "`pip install scyjava`",
}


OPENSLIDE_FORMATS = {
    "svs",
    "dcm",
    "ndpi",
    "vms",
    "vmu",
    "scn",
    "mrxs",
    "svslide",
    "bif",
    "tif",
    "tiff",
}

TIFFSLIDE_FORMATS = {
    "svs",
    "ndpi",
    "vms",
    "vmu",
    "scn",
    "tif",
    "tiff",
}

CUCIM_FORMATS = {
    "svs",
    "tiff",
}

# Bio-Formats is the last resort for all formats


def get_reader(reader: str = None, format: str = None):
    """
    Get the reader based on the input format and reader name.

    Parameters
    ----------
    reader : str
        Name of the reader to use. If None, will try to find a suitable reader.
    format : str
        Format of the image file. If None, will try to find a suitable reader.

    Returns
    -------
    reader : ReaderBase
        Reader object to read the image file.

    """
    readers = {r: None for r in Reader}
    error_stack = {r: None for r in Reader}
    catch_error = (ModuleNotFoundError, OSError, ImportError)

    try:
        import openslide

        readers[Reader.OPENSLIDE] = OpenSlideReader
    except catch_error as e:
        error_stack[Reader.OPENSLIDE] = e

    try:
        import tiffslide

        readers[Reader.TIFFSLIDE] = TiffSlideReader
    except catch_error as e:
        error_stack[Reader.TIFFSLIDE] = e

    try:
        import scyjava

        readers[Reader.BIOFORMATS] = BioFormatsReader
    except catch_error as e:
        error_stack[Reader.BIOFORMATS] = e

    try:
        import cucim

        readers[Reader.CUCIM] = CuCIMReader
    except catch_error as e:
        error_stack[Reader.CUCIM] = e

    reader_candidates = []
    if format is not None and reader is None:
        format = format.strip().lstrip(".").lower()
        if format in OPENSLIDE_FORMATS:
            reader_candidates.append(Reader.OPENSLIDE)
        if format in TIFFSLIDE_FORMATS:
            reader_candidates.append(Reader.TIFFSLIDE)
        reader_candidates.append(Reader.BIOFORMATS)
        if format in CUCIM_FORMATS:
            reader_candidates.append(Reader.CUCIM)
    else:
        reader_candidates = list(Reader)

    if reader is None:
        for i in reader_candidates:
            reader = readers.get(i)
            if reader is not None:
                return reader
        msg = "None of the readers are available:"
        for i in reader_candidates:
            msg += f"\n{i}: {error_stack[i]}"
        raise ValueError(msg)
    else:
        reader = Reader(reader)
        if reader not in reader_candidates:
            raise ValueError(
                f"Requested reader not available, "
                f"must be one of {reader_candidates}"
            )
        else:
            used_reader = readers.get(reader)
            if used_reader is None:
                raise ValueError(
                    f"Requested reader not available: {reader}, if you don't have it installed, "
                    f"{InstallationTips[reader]}\n"
                    f"Following error occurred: {error_stack[reader]}"
                )
            return used_reader



import json
from dataclasses import dataclass, asdict
from typing import Optional


@dataclass
class TileSpec:
    """Data class for storing tile specifications.

    Parameters
    ----------
    height : int
        The height of the tile.
    width : int
        The width of the tile.
    stride_height : int
        The height of the stride.
    stride_width : int
        The width of the stride.
    raw_{height, width} : int
        The height/width of the tile when retrieving images.
    raw_stride_{height, width}: int
        The height/width of the stride when retrieving images.
    tissue_name : str
        The name of the tissue.
    level : int, default: 0
        The level of the tile.
    downsample : float, default: 1
        The downsample factor.
    mpp : float, default: None
        The microns per pixel.

    """

    height: int
    width: int
    stride_height: int
    stride_width: int
    raw_height: int
    raw_width: int
    raw_stride_height: int
    raw_stride_width: int
    tissue_name: str
    level: int = 0
    downsample: float = 1
    mpp: Optional[float] = None

    def to_dict(self):
        return asdict(self)

    def to_json(self):
        return json.dumps(asdict(self))



from __future__ import annotations

import json
from functools import singledispatch
from typing import Optional, List, Mapping
from dataclasses import dataclass, asdict, field

import cv2
import numpy as np
from PIL import Image

# AnnData cannot serialize Tuple
SHAPE = List[int]


@dataclass
class SlideProperties:
    """
    The properties of the slide

    Attributes
    ----------
    shape : [height, width]
        The shape of the slide
    n_level : int
        The number of pyramids levels
    level_shape : List[[height, width]]
        The shape of each pyramid level
    level_downsample : List[float]
        The downsample factor of each pyramid level
    mpp : Optional[float]
        The physical size of each pixel in microns
    magnification : Optional[float]
        The magnification of the slide
    bounds : Optional[SHAPE]
        The bounds of the slide, in the format [x, y, height, width]
        This is the region of the slide that contains tissue
    raw : Optional[str]
        The raw metadata in serialized json format

    """

    shape: SHAPE
    n_level: int
    level_shape: List[SHAPE]
    level_downsample: List[float]
    mpp: Optional[float] = None
    magnification: Optional[float] = None
    bounds: Optional[SHAPE] = None
    raw: Optional[str] = field(default=None, repr=False, compare=False)

    @classmethod
    def from_mapping(self, metadata: Mapping):
        """Create SlideProperties from a mapping"""
        metadata = parse_metadata(metadata)
        return SlideProperties(**metadata)

    def to_dict(self):
        """Convert the properties to a dictionary"""
        return asdict(self)

    def to_json(self):
        """Convert the properties to a json string"""
        return json.dumps(asdict(self))

    def _repr_html_(self):
        rows = []
        for k, v in self.to_dict().items():
            if k != "raw":
                rows.append(f"<tr><td>{k}</td><td>{v}</td></tr>")

        return (
            "<h4>Slide Properties</h4><table><tr><th>Field</th><th>Value</th></tr>"
            + "".join(rows)
            + "</table>"
        )


class ReaderBase:
    """The base class for all reader

    This class defines the basic interface for all reader

    Attributes
    ----------
    file : str
        The path to the image file
    properties : :class:`SlideProperties <wsi.reader.SlideProperties>`
        The properties of the slide
    name : str
        The name of the reader
    reader : Any
        The reader object

    """

    file: str
    properties: SlideProperties
    name = "base"
    _reader = None

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.file}')"

    def __del__(self):
        self.detach_reader()

    def translate_level(self, level):
        """Translate the level to the actual level

        level -1 refer to the lowest resolution level

        Parameters
        ----------
        level : int
            The level to translate

        """
        levels = np.arange(self.properties.n_level)
        if level >= len(levels):
            raise ValueError(f"Request level {level} not exist")
        return levels[level]

    def get_region(self, x, y, width, height, level=0, **kwargs):
        """Get a region from image with top-left corner
        This should return a numpy array in xyc format
        """
        raise NotImplementedError

    def get_center(self, x, y, width, height, level=0, **kwargs):
        """Get a patch from image with center"""
        x -= width / 2
        y -= height / 2
        return self.get_region(x, y, width, height, level=level, **kwargs)

    def get_thumbnail(self, size, **kwargs):
        """Get a thumbnail of the image"""
        raise NotImplementedError

    def get_level(self, level) -> np.ndarray[np.uint8]:
        """
        Extract the image at the given level

        A very expensive operation, use with caution

        Parameters
        ----------
        level : int
            The level to extract

        """
        height, width = self.properties.level_shape[level]
        return self.get_region(0, 0, width, height, level=level)

    def set_properties(self, properties: SlideProperties | Mapping):
        if isinstance(properties, SlideProperties):
            self.properties = properties
        else:
            self.properties = SlideProperties.from_mapping(properties)

    def create_reader(self):
        """Create the reader

        This function should be implemented in the subclass

        1. create the reader
        2. assign the reader to self._reader

        """
        # The basic fallback implementation to create reader
        raise NotImplementedError

    def detach_reader(self):
        """Detach the reader

        This function will be called when the object is deleted or
        sent to other process

        Please implement this function in the subclass

        1. Close the reader
        2. Set self._reader to None
        3. Clean up any resources that are not python-managed

        """
        raise NotImplementedError

    @staticmethod
    def resize_img(
        img: np.ndarray,
        dsize: SHAPE = None,
        scale: float = None,
    ):
        dim = np.asarray(img.shape)
        if dsize is not None:
            return cv2.resize(img, dsize)
        if scale is not None:
            dim = np.array(dim * scale, dtype=int)
            return cv2.resize(img, dim)

    @property
    def reader(self):
        if self._reader is None:
            self.create_reader()
        return self._reader


@singledispatch
def convert_image(img):
    raise NotImplementedError(f"Unsupported type {type(img)}")


@convert_image.register(Image.Image)
def _(img: Image.Image):
    return cv2.cvtColor(np.asarray(img), cv2.COLOR_RGBA2RGB).astype(np.uint8)


@convert_image.register(np.ndarray)
def _(img: np.ndarray):
    return cv2.cvtColor(img, cv2.COLOR_RGBA2RGB).astype(np.uint8)


MAG_KEY = "objective-power"
MPP_KEYS = ("mpp-x", "mpp-y")

LEVEL_HEIGHT_KEY = lambda level: f"level[{level}].height"  # noqa: E731
LEVEL_WIDTH_KEY = lambda level: f"level[{level}].width"  # noqa: E731
LEVEL_DOWNSAMPLE_KEY = lambda level: f"level[{level}].downsample"  # noqa: E731


def parse_metadata(metadata: Mapping):
    metadata = dict(metadata)
    new_metadata = {}
    for k, v in metadata.items():
        new_metadata[".".join(k.split(".")[1::])] = v
    metadata.update(new_metadata)

    fields = set(metadata.keys())

    mpp_keys = []
    # openslide specific mpp keys
    if MPP_KEYS[0] in fields:
        mpp_keys.append(MPP_KEYS[0])
    if MPP_KEYS[1] in fields:
        mpp_keys.append(MPP_KEYS[1])
    for k in fields:
        # Any keys end with .mpp
        if k.lower().endswith("mpp"):
            mpp_keys.append(k)

    mpp = None
    for k in mpp_keys:
        mpp_tmp = metadata.get(k)
        if mpp_tmp is not None:
            mpp = float(mpp_tmp)

    # search magnification
    # search other available mpp keys
    mag_keys = []
    if MAG_KEY in fields:
        mag_keys.append(MAG_KEY)
    for k in fields:
        # Any keys end with .mpp
        if k.lower().endswith("appmag"):
            mag_keys.append(k)

    mag = None
    for k in mag_keys:
        mag_tmp = metadata.get(k)
        if mag_tmp is not None:
            mag = float(mag_tmp)

    level_shape = []
    level_downsample = []

    # Get the number of levels
    n_level = 0
    while f"level[{n_level}].width" in fields:
        n_level += 1

    for level in range(n_level):
        height = metadata.get(LEVEL_HEIGHT_KEY(level))
        width = metadata.get(LEVEL_WIDTH_KEY(level))
        downsample = metadata.get(LEVEL_DOWNSAMPLE_KEY(level))

        level_shape.append((int(height), int(width)))

        if downsample is not None:
            downsample = float(downsample)
        level_downsample.append(downsample)

    shape = level_shape[0]

    # Get the bounds
    bounds_x = metadata.get("bounds-x")
    bounds_y = metadata.get("bounds-y")
    bounds_height = metadata.get("bounds-height")
    bounds_width = metadata.get("bounds-width")
    # The key may still exist but the value is None
    # So we need to check if the value is None
    bounds_x = 0 if bounds_x is None else int(bounds_x)
    bounds_y = 0 if bounds_y is None else int(bounds_y)
    bounds_height = shape[0] if bounds_height is None else int(bounds_height)
    bounds_width = shape[1] if bounds_width is None else int(bounds_width)
    bounds = [bounds_x, bounds_y, bounds_height, bounds_width]

    metadata = {
        "mpp": mpp,
        "magnification": mag,
        "shape": list(shape),
        "n_level": n_level,
        "level_shape": [list(i) for i in level_shape],
        "level_downsample": level_downsample,
        "raw": json.dumps(metadata),
        "bounds": bounds,
    }

    return metadata



try:
    import torch
    import torchstain.torch.normalizers as norm
    from torchvision.transforms.v2 import Compose, ToImage, ToDtype, Lambda

    class ColorNormalizer(torch.nn.Module):
        T = Compose(
            [ToImage(), ToDtype(torch.float32, scale=True), Lambda(lambda x: x * 255)]
        )

        def __init__(self, method="macenko"):
            super().__init__()

            self.method = method
            if method == "macenko":
                normalizer = norm.TorchMacenkoNormalizer()
            elif method == "reinhard":
                normalizer = norm.TorchReinhardNormalizer()
                normalizer.target_means = torch.tensor([72.909996, 20.8268, -4.9465137])
                normalizer.target_stds = torch.tensor([18.560713, 14.889295, 5.6756697])
            elif method == "reinhard_modified":
                normalizer = norm.TorchReinhardNormalizer(method="modified")
                normalizer.target_means = torch.tensor([72.909996, 20.8268, -4.9465137])
                normalizer.target_stds = torch.tensor([18.560713, 14.889295, 5.6756697])
            else:
                raise NotImplementedError(
                    f"Requested method '{method}' not implemented"
                )
            self.normalizer = normalizer

        def __repr__(self):
            return f"ColorNormalizer(method='{self.method}')"

        def forward(self, img):
            t_img = self.T(img)
            if self.method == "macenko":
                norm, _, _ = self.normalizer.normalize(I=t_img)
            else:
                norm = self.normalizer.normalize(I=t_img)
            return norm
except (ImportError, ModuleNotFoundError):

    class ColorNormalizer:
        def __init__(self, *args, **kwargs):
            raise ImportError(
                "To use color normalization, please install torch, torchvision and torchstain."
            )



from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import numpy as np
from anndata import AnnData
from fsspec.core import url_to_fs
from rich.progress import track
from spatialdata import read_zarr, SpatialData
from spatialdata.models import Image2DModel
from spatialdata.transformations import Scale

from ._download import CacheDownloader
from .._model import WSIData
from ..reader import get_reader, to_datatree


def open_wsi(
    wsi,
    store=None,
    reader=None,
    download=True,
    name=None,
    cache_dir=None,
    pbar=True,
    attach_images=False,
    image_key="wsi",
    save_images=False,
    attach_thumbnail=True,
    thumbnail_key="wsi_thumbnail",
    thumbnail_size=2000,
    save_thumbnail=True,
    **kwargs,
):
    """Open a whole slide image.

    You can open a whole slide image from a URL or a local file.
    If load from remote URL, the image will be downloaded and cached (default to current the working directory).
    You can also attach images and thumbnail to the SpatialData object. By default, only the thumbnail is attached,
    the thumbnail is a downsampled version of the whole slide image, the original image is not attached to save disk space
    when you save the WSIData object on disk.

    Parameters
    ----------
    wsi : str or Path
        The URL to whole slide image.
    store : str, optional
        The backed file path, by default will create
        a zarr file with the same name as the slide file.
        You can either supply a file path or a directory.
        If a directory is supplied, the zarr file will be created in that directory.
        This is useful when you want to store all zarr files in a specific location.
    reader : str, optional
        Reader to use, by default "auto", choosing available reader, first openslide, then tifffile.
    download : bool, optional
        Whether to download the slide.
    name : str, optional
        The name of the slide.
    cache_dir : str, optional
        The cache directory, by default file will be stored in working direction.
    pbar : bool, optional
        Whether to show progress bar, by default True.
    attach_images : bool, optional, default: False
        Whether to attach whole slide image to image slot in the spatial data object.
    image_key : str, optional
        The key to store the whole slide image, by default "wsi_thumbnail".
    save_images : bool, optional, default: False
        Whether to save the whole slide image to on the disk.
        Only works for wsi.save() method.
    attach_thumbnail : bool, optional, default: True
        Whether to attach thumbnail to image slot in the spatial data object.
    thumbnail_key : str, optional
        The key to store the thumbnail, by default "wsi_thumbnail".
    thumbnail_size : int, optional, default: 2000
        The size of the thumbnail.
    save_thumbnail : bool, optional, default: True
        Whether to save the thumbnail to on the disk.

        Only works for wsi.save() method.

    Returns
    -------
    :class:`WSIData`
        Whole slide image data.

    Examples
    --------

    .. code-block:: python

        >>> from wsidata import open_wsi
        >>> wsi = open_wsi("https://bit.ly/3ZvbzVc")

    """

    # Check if the slide is a file or URL
    wsi = str(wsi)
    fs, wsi_path = url_to_fs(wsi)
    if not fs.exists(wsi_path):
        raise ValueError(f"Slide {wsi} not existed or not accessible.")

    # Early attempt with reader
    format = Path(wsi).suffix
    ReaderCls = get_reader(reader, format=format)

    if download and fs.protocol != "file":
        downloader = CacheDownloader(wsi_path, name=name, cache_dir=cache_dir)
        wsi = downloader.download(pbar)

    reader_obj = ReaderCls(wsi)
    wsi = Path(wsi)
    if store is None:
        store = wsi.with_suffix(".zarr")
    else:
        # We also support write all backed file to a directory
        backed_file_p = Path(store)
        if backed_file_p.is_dir():
            if is_zarr_dir(backed_file_p):
                store = backed_file_p
            else:
                zarr_name = Path(wsi).with_suffix(".zarr").name
                store = backed_file_p / zarr_name
        else:
            store = backed_file_p

    if store.exists():
        sdata = read_zarr(store)
    else:
        sdata = SpatialData()

    exclude_elements = []

    if attach_images and image_key not in sdata:
        images_datatree = to_datatree(reader_obj)
        sdata.images[image_key] = images_datatree
        if not save_images:
            exclude_elements.append(image_key)

    if attach_thumbnail and thumbnail_key not in sdata:
        thumbnail = reader_obj.get_thumbnail(thumbnail_size)
        thumbnail_shape = thumbnail.shape
        origin_shape = reader_obj.properties.shape
        scale_x, scale_y = (
            origin_shape[0] / thumbnail_shape[0],
            origin_shape[1] / thumbnail_shape[1],
        )

        if thumbnail is not None:
            sdata.images[thumbnail_key] = Image2DModel.parse(
                np.asarray(thumbnail).transpose(2, 0, 1),
                dims=("c", "y", "x"),
                transformations={"global": Scale([scale_x, scale_y], axes=("x", "y"))},
            )
            if not save_thumbnail:
                exclude_elements.append(thumbnail_key)

    slide_data = WSIData.from_spatialdata(sdata, reader_obj)
    slide_data.set_exclude_elements(exclude_elements)
    slide_data.set_wsi_store(store)
    return slide_data


def agg_wsi(
    slides_table,
    feature_key,
    tile_key="tiles",
    agg_key="agg",
    wsi_col=None,
    store_col=None,
    error="raise",
):
    """
    Aggregate feature from a whole slide image.

    Parameters
    ----------
    slides_table: pd.DataFrame
        The table of slides, including information of whole slide image and .zarr paths and metadata.

        Backed file path to the anndata file is optional, by default the same directory as the
        whole slide image.
    feature_key: str
        The feature key on which aggregation should be run on.
    tile_key: str
        The tile key.
    agg_key: str
        The output aggregation key in the varm slot.
    wsi_col: str
        The column name of the whole slide image paths.
    store_col: str
        The column name of the backed file.
    error: str
        Whether to raise error when file not existed.

    Returns
    -------
    AnnData
        The aggregated feature space.
    """
    if wsi_col is None and store_col is None:
        raise ValueError("Either wsi_col or backed_file_col must be provided.")

    if store_col is not None:
        backed_files = slides_table[store_col]
    elif wsi_col is not None:
        backed_files = slides_table[wsi_col].apply(
            lambda x: Path(x).with_suffix(".zarr")
        )

    jobs = []
    with ThreadPoolExecutor() as executor:
        for backed_f in backed_files:
            job = executor.submit(
                _agg_wsi, backed_f, feature_key, tile_key, agg_key, error
            )
            jobs.append(job)

    results = []
    for job in track(
        jobs,
        description=f"Aggregation of {len(jobs)} slides",
    ):
        results.append(job.result())

    if error != "raise":
        mask = np.asarray([r is not None for r in results])
        X = np.vstack(np.asarray(results, dtype=object)[mask])
        slides_table = slides_table[mask]

    else:
        X = np.vstack(results)

    # Convert index to string
    slides_table.index = slides_table.index.astype(str)
    return AnnData(X, obs=slides_table)


def _agg_wsi(f, feature_key, tile_key, agg_key, error="raise"):
    """
    Aggregate feature from a whole slide image.

    Parameters
    ----------
    f: str
        The backed file path to the anndata file stored as .zarr.
    feature_key: str
        The feature key on which aggregation should be run on.
    tile_key: str
        The tile key.
    agg_key:
        The output aggregation key in the varm slot.
    error: str
        Whether to raise error when file not existed.

    Returns
    -------
    np.ndarray
        The aggregated feature space.

    """
    if not Path(f).exists():
        if error == "raise":
            raise ValueError(f"File {f} not existed.")
        else:
            return None
    try:
        import zarr
        from anndata import read_zarr

        tables = zarr.open(f"{f}/tables")
        available_keys = list(tables.keys())
        if feature_key not in available_keys:
            feature_key = f"{feature_key}_{tile_key}"
        s = read_zarr(f"{f}/tables/{feature_key}")
        return s.varm[agg_key]
    except Exception as e:
        if error == "raise":
            raise e
        else:
            return None


def is_zarr_dir(path):
    """
    Detect if the given directory is a Zarr storage using the Zarr library.

    Parameters:
        path (str): The path to the directory.

    Returns:
        bool: True if the directory is a Zarr storage, False otherwise.
    """
    import zarr

    try:
        zarr.open_group(path, mode="r")
        return True
    except Exception:
        return False



from __future__ import annotations

import warnings
from functools import cached_property
from pathlib import Path
from typing import Literal, Generator

import numpy as np
from PIL.Image import Image, fromarray
from anndata import AnnData
from spatialdata import SpatialData
from spatialdata.models import SpatialElement

from .tilespec import TileSpec
from .._accessors import FetchAccessor, IterAccessor, DatasetAccessor
from .._utils import find_stack_level
from ..reader import ReaderBase


class WSIData(SpatialData):
    """
    A container class combining :class:`SpatialData <spatialdata.SpatialData>`
    and a whole slide image reader.

    .. note::
     Use the :func:`open_wsi` function to create a WSIData object.

    By default, the whole slide image is not attached to the SpatialData.
    A thumbnail version of the whole slide image is attached for visualization purpose.

    The WSIData contains four main components:

    .. list-table::
     :header-rows: 1

     * -
       - Whole slide image
       - Tissue contours
       - Tile locations
       - Features

     * - **SpatialData Slot**
       - :bdg-danger:`images`
       - :bdg-danger:`shapes`
       - :bdg-danger:`shapes`
       - :bdg-danger:`tables`

     * - **Default Key**
       - :bdg-info:`wsi_thumbnail`
       - :bdg-info:`tissues`
       - :bdg-info:`tiles`
       - :bdg-info:`\{feature_key\}_\{tile_key\}`

     * - | **Attributes**
         | **Slot**
         | **Key**
       - | :class:`SlideProperties <wsidata.reader.SlideProperties>`
         | :bdg-danger:`tables`
         | :bdg-info:`slide_properties`
       -
       - | :class:`TileSpec <wsidata.TileSpec>`
         | :bdg-danger:`tables`
         | :bdg-info:`tile_spec`
       -

     * - **Content**
       - | :class:`DataArray <xarray.DataArray>`
         | (c, y, x) format.
       - | :class:`GeoDataFrame <geopandas.GeoDataFrame>` with columns:
         | :bdg-black:`tissue_id`
         | :bdg-black:`geometry`
       - | :class:`GeoDataFrame <geopandas.GeoDataFrame>` with columns:
         | :bdg-black:`tile_id`
         | :bdg-black:`x`, :bdg-black:`y`
         | :bdg-black:`tissue_id`
         | :bdg-black:`geometry`
       - | :class:`AnnData <anndata.AnnData>` with:
         | :code:`X`: The feature matrix
         | :code:`varm`: :bdg-black:`agg_slide`, :bdg-black:`agg_tissue`



     You can interact with WSIData using the following accessors:

     - :class:`get <wsidata.GetAccessor>`: Access data from the WSIData object.
     - :class:`iter <wsidata.IterAccessor>`: Iterate over data in the WSIData object.
     - :class:`ds <wsidata.DatasetAccessor>`: Create deep learning datasets from the WSIData object.
     - To implement your own accessors, use :func:`register_wsidata_accessor <wsidata.register_wsidata_accessor>`.

     For analysis purpose, you can override two slide properties:

     - microns per pixel (mpp): Using the :meth:`set_mpp` method.
     - bounds: Using the :meth:`set_bounds` method.

     Parameters
     ----------
     reader : :class:`ReaderBase <wsidata.reader.ReaderBase>`
         A reader object that can interface with the whole slide image file.
     sdata : :class:`SpatialData <spatialdata.SpatialData>`
         A SpatialData object for storing analysis data.
     backed_file : str or Path
         Storage location to the SpatialData object.
     slide_properties_source : {'slide', 'sdata'}, default: 'sdata'
         The source of the slide properties.

         - "slide": load from the reader object.
         - "sdata": load from the SpatialData object.

     Attributes
     ----------
     properties : :class:`SlideProperties <wsidata.reader.SlideProperties>`
         The properties of the whole slide image.
     reader : :class:`ReaderBase <wsidata.reader.ReaderBase>`
         The reader object for interfacing with the whole slide image.
     sdata : :class:`SpatialData <spatialdata.SpatialData>`
         The SpatialData object containing the spatial data.
     backed_file : Path
         The path to the backed file.

    """

    TILE_SPEC_KEY = "tile_spec"
    SLIDE_PROPERTIES_KEY = "slide_properties"

    @classmethod
    def from_spatialdata(cls, sdata, reader=None, **kws):
        d = cls(
            images=sdata.images,
            labels=sdata.labels,
            shapes=sdata.shapes,
            tables=sdata.tables,
            points=sdata.points,
            reader=reader,
            **kws,
        )
        d.path = sdata.path
        return d

    def __init__(
        self,
        images=None,
        labels=None,
        shapes=None,
        tables=None,
        points=None,
        reader: ReaderBase = None,
        slide_properties_source: Literal["slide", "sdata"] = "sdata",
    ):
        self._reader = reader
        self._wsi_store = None
        self.slide_properties_source = slide_properties_source
        self._exclude_elements = set()
        super().__init__(
            images=images,
            labels=labels,
            shapes=shapes,
            tables=tables,
            points=points,
        )

        if self.SLIDE_PROPERTIES_KEY not in self:
            self.tables[self.SLIDE_PROPERTIES_KEY] = AnnData(
                uns=reader.properties.to_dict()
            )
        else:
            # Try to load the slide properties from the spatial data
            if slide_properties_source == "slide":
                reader_properties = self.tables[self.SLIDE_PROPERTIES_KEY].uns
                if reader_properties != reader.properties.to_dict():
                    # Update the reader properties
                    reader.properties.from_mapping(reader_properties)
                    warnings.warn(
                        "Slide properties in the spatial data is different from the reader properties.",
                        UserWarning,
                        stacklevel=find_stack_level(),
                    )

    def __repr__(self):
        return (
            f"WSI: {self.reader.file}\n"
            f"Reader: {self.reader.name}\n"
            f"{super().__repr__()}"
        )

    def set_exclude_elements(self, elements):
        self._exclude_elements.update(elements)

    def set_wsi_store(self, store: str | Path):
        self._wsi_store = Path(store)

    def _gen_elements(
        self, include_table: bool = False
    ) -> Generator[tuple[str, str, SpatialElement | AnnData], None, None]:
        for i in super()._gen_elements(include_table):
            if i[1] not in self._exclude_elements:
                yield i

    def close(self):
        """Close the reader object."""
        self.reader.detach_reader()

    @property
    def reader(self):
        return self._reader

    @property
    def properties(self):
        return self.reader.properties

    @property
    def wsi_store(self):
        return self._wsi_store

    @property
    def thumbnail(self):
        return self.get_thumbnail(size=500, as_array=False)

    @property
    def name(self):
        return Path(self.reader.file).name

    def tile_spec(self, key: str) -> TileSpec:
        """
        Get the :class:`TileSpec` for a collection of tiles.

        Parameters
        ----------
        key : str
            The key of the tiles.

        """
        if self.TILE_SPEC_KEY in self:
            spec = self.tables[self.TILE_SPEC_KEY].uns[key]
            return TileSpec(**spec)

    def set_mpp(self, mpp):
        """Set the microns per pixel (mpp) of the whole slide image."""
        self.properties.mpp = mpp
        self.tables[self.SLIDE_PROPERTIES_KEY].uns["mpp"] = mpp

    def set_bounds(self, bounds):
        """Set the bounds of the whole slide image."""
        self.properties.bounds = bounds
        self.tables[self.SLIDE_PROPERTIES_KEY].uns["bounds"] = bounds

    def read_region(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        level: int = 0,
        **kwargs,
    ) -> np.ndarray[np.uint8]:
        """Read a region from the whole slide image.

        Parameters
        ----------
        x : int
            The x-coordinate at level 0.
        y : int
            The y-coordinate at level 0.
        width : int
            The width of the region in pixels.
        height : int
            The height of the region in pixels.
        level : int, default: 0
            The pyramid level.

        """
        if self.reader is None:
            raise ValueError("The reader object is not attached.")
        return self.reader.get_region(x, y, width, height, level=level, **kwargs)

    def get_thumbnail(self, size=500, as_array=False) -> np.ndarray[np.uint8] | Image:
        """Get the thumbnail of the whole slide image.

        Parameters
        ----------
        as_array : bool, default: False
            Return as numpy array.

        """
        img = self.reader.get_thumbnail(size=size)
        if as_array:
            return img
        else:
            return fromarray(img)

    def write(
        self,
        file_path=None,
        overwrite: bool = True,
        consolidate_metadata: bool = True,
        format=None,
    ):
        if file_path is not None:
            file_path = Path(file_path)
            if self.path is None:
                self.path = file_path
        else:
            file_path = self._wsi_store
        super().write(
            file_path=file_path,
            overwrite=overwrite,
            consolidate_metadata=consolidate_metadata,
            format=format,
        )

    # def save(self, file=None, consolidate_metadata: bool = True):
    #     # Create the store first
    #     if file is not None:
    #         file = Path(file)
    #         if self.path is None:
    #             self.path = file
    #     else:
    #         file = self._wsi_store
    #     store = parse_url(file, mode="w").store
    #     self.path = file
    #     _ = zarr.group(store=store, overwrite=False)
    #     store.close()
    #     # WARNING: This is not thread-safe, data may be corrupted if multiple threads write to the same file
    #     # Waiting for SpatialData to support thread-safe writing
    #     for elem_type, elem_key, elem in self._gen_elements(include_table=True):
    #         self._write_element(
    #             element=elem,
    #             zarr_container_path=file,
    #             element_type=elem_type,
    #             element_name=elem_key,
    #             overwrite=True,
    #         )
    #     if consolidate_metadata:
    #         self.write_consolidated_metadata()

    def _check_feature_key(self, feature_key, tile_key=None):
        msg = f"{feature_key} doesn't exist"
        if feature_key in self:
            return feature_key
        else:
            if tile_key is not None:
                feature_key = f"{feature_key}_{tile_key}"
                if feature_key in self:
                    return feature_key
                msg = f"Neither {feature_key} or {feature_key}_{tile_key} exist"

        raise KeyError(msg)

    # Define default accessors as property,
    # this will enable auto-completion in IDE
    # For extension, user can define their own accessors
    # using register_wsidata_accessor
    @cached_property
    def fetch(self):
        return FetchAccessor(self)

    @cached_property
    def iter(self):
        return IterAccessor(self)

    @cached_property
    def ds(self):
        return DatasetAccessor(self)



from __future__ import annotations

from typing import NamedTuple, Sequence, Dict, List, Tuple

import cv2
import numpy as np
from shapely import MultiPolygon, Polygon, box, clip_by_rect
from shapely.affinity import translate, scale

from .._normalizer import ColorNormalizer


class TissueContour(NamedTuple):
    tissue_id: int
    contour: np.ndarray
    holes: List[np.ndarray | Polygon] | None = None


class TissueImage(NamedTuple):
    tissue_id: int
    x: int
    y: int
    image: np.ndarray
    mask: np.ndarray


class TileImage(NamedTuple):
    id: int
    x: int
    y: int
    tissue_id: int
    image: np.ndarray
    anno_mask: np.ndarray | None = None
    anno_shapes: List[Tuple[Polygon, str, int]] | None = None

    def __repr__(self):
        image_dtype = self.image.dtype
        if self.anno_mask is not None:
            mask_repr = (
                f"(shape: {self.anno_mask.shape}, dtype: {self.anno_mask.dtype})"
            )
        else:
            mask_repr = None

        return (
            f"TileImage(id={self.id}, "
            f"x={self.x}, y={self.y}, "
            f"tissue_id={self.tissue_id}, "
            f"image=(shape: {self.image.shape}, dtype: {image_dtype}), "
            f"anno_mask={mask_repr}, "
            f"anno_shapes=({len(self.anno_shapes)} shapes))"
        )


class IterAccessor(object):
    """An accessor to iterate over the WSI data.

    Usage: `wsidata.iter`

    """

    def __init__(self, obj):
        self._obj = obj

    def tissue_contours(
        self,
        key,
        as_array: bool = False,
        dtype: np.dtype = None,
        shuffle: bool = False,
        seed: int = 0,
    ) -> TissueContour:
        """A generator to extract tissue contours from the WSI.

        Parameters
        ----------
        key : str
            The tissue key.
        as_array : bool, default: False
            Return the contour as an array.
            If False, the contour is returned as a shapely geometry.
        dtype : np.dtype, default: None
            The data type of the array if as_array is True.
        shuffle : bool, default: False
            If True, return tissue contour in random order.
        seed : int, default: 0

        Returns
        -------
        TissueContour
            A named tuple with fields:

            - tissue_id : The tissue id.
            - contour : The tissue contour as a shapely geometry or an array.
            - holes : The holes in the tissue contour.

        """

        contours = self._obj.shapes[key]

        if shuffle:
            contours = contours.sample(frac=1, random_state=seed)

        for ix, cnt in contours.iterrows():
            tissue_id = cnt["tissue_id"]
            if as_array:
                yield TissueContour(
                    tissue_id=tissue_id,
                    contour=np.array(cnt.geometry.exterior.coords, dtype=dtype),
                    holes=[
                        np.asarray(h.coords, dtype=dtype)
                        for h in cnt.geometry.interiors
                    ],
                )
            else:
                yield TissueContour(
                    tissue_id=tissue_id,
                    contour=cnt.geometry,
                    holes=[Polygon(i) for i in cnt.geometry.interiors],
                )

    def tissue_images(
        self,
        key,
        level=0,
        mask_bg=False,
        tissue_mask=False,
        color_norm: str = None,
        format: str = "xyc",
    ) -> TissueImage:
        """Extract tissue images from the WSI.

        Parameters
        ----------
        key : str
            The tissue key.
        level : int, default: 0
            The level to extract the tissue images.
        mask_bg : bool | int, default: False
            Mask the background with the given value.

            If False, the background is not masked.

            If True, the background is masked with 0.

            If an integer, the background is masked with the given value.
        color_norm : str, {"macenko", "reinhard"}, default: None
            Color normalization method.
        format : str, {"xyc", "cyx"}, default: "xyc"
            The channel format of the image.

        Returns
        -------
        TissueImage
            A named tuple with fields:

            - tissue_id : The tissue id.
            - x : The x-coordinate of the image.
            - y : The y-coordinate of the image.
            - image : The tissue image.
            - mask : The tissue mask.

        """
        import cv2

        level_downsample = self._obj.properties.level_downsample[level]
        if color_norm is not None:
            cn = ColorNormalizer(method=color_norm)
            cn_func = lambda x: cn(x).numpy().astype(np.uint8)  # noqa
        else:
            cn_func = lambda x: x  # noqa

        if isinstance(mask_bg, bool):
            do_mask = mask_bg
            if do_mask:
                mask_bg = 0
        else:
            do_mask = True
            mask_bg = mask_bg
        for tissue_contour in self.tissue_contours(key):
            ix = tissue_contour.tissue_id
            contour = tissue_contour.contour
            holes = [np.asarray(h.coords) for h in contour.interiors]
            minx, miny, maxx, maxy = contour.bounds
            x = int(minx)
            y = int(miny)
            w = int(maxx - minx) / level_downsample
            h = int(maxy - miny) / level_downsample
            img = self._obj.reader.get_region(x, y, w, h, level=level)
            img = cn_func(img)

            mask = None
            if do_mask or tissue_mask:
                mask = np.zeros_like(img[:, :, 0])
                # Offset and scale the contour
                offset_x, offset_y = x / level_downsample, y / level_downsample
                coords = np.array(contour.exterior.coords) - [offset_x, offset_y]
                coords = (coords / level_downsample).astype(np.int32)
                # Fill the contour with 1
                cv2.fillPoly(mask, [coords], 1)

                # Fill the holes with 0
                for hole in holes:
                    hole -= [offset_x, offset_y]
                    hole = (hole / level_downsample).astype(np.int32)
                    cv2.fillPoly(mask, [hole], 0)

            if do_mask:
                # Fill everything that is not the contour
                # (which is background) with 0
                img[mask != 1] = mask_bg
            if not tissue_mask:
                mask = None
            else:
                mask = mask.astype(bool)

            if format == "cyx":
                img = img.transpose(2, 0, 1)

            yield TissueImage(tissue_id=ix, x=x, y=y, image=img, mask=mask)

    def tile_images(
        self,
        key,
        raw=False,
        color_norm: str = None,
        format: str = "xyc",
        annotation_key: str = None,
        annotation_name: str | Sequence[str] = None,
        annotation_label: str | Dict[str, int] = None,
        mask_dtype: np.dtype = None,
        shuffle: bool = False,
        sample_n: int = None,
        seed: int = 0,
    ) -> TileImage:
        """Extract tile images from the WSI.

        Parameters
        ----------
        key : str
            The tile key.
        raw : bool, default: True
            Return the raw image without resizing.
            If False, the image is resized to the requested tile size.
        color_norm : str, {"macenko", "reinhard"}, default: None
            Color normalization method.
        format : str, {"xyc", "cyx"}, default: "xyc"
            The channel format of the image.
        annotation_key : str, default: None
            The key to the annotation table in :bdg-danger:`shapes` slot.
        annotation_name : str or array of str, default: None
            The name of the annotation column in the annotation table or a list of names.
        annotation_label : str or dict, default: None
            The name of the label column in the annotation table or a dictionary of label mappings.
        mask_dtype : np.dtype, default: np.uint8
            The data type of the annotation mask, if you have annotation labels more than 255,
            consider using data type with higher bound.
        shuffle : bool, default: False
            If True, return tile images in random order.
        sample_n : int, default: None
            The number of samples to return.
        seed : int, default: 0
            The random seed.

        Returns
        -------
        TileImage
            A named tuple with fields:

            - id: The tile id.
            - x: The x-coordinate of the tile.
            - y: The y-coordinate of the tile.
            - tissue_id: The tissue id.
            - image: The tile image.
            - anno_mask: The annotation mask.
            - anno_shapes: The annotation shapes.

        """
        tile_spec = self._obj.tile_spec(key)

        if color_norm is not None:
            cn = ColorNormalizer(method=color_norm)
            cn_func = lambda x: cn(x).numpy().astype(np.uint8)  # noqa
        else:
            cn_func = lambda x: x  # noqa

        create_anno_mask = False
        anno_tb = None
        mask_size = (
            (tile_spec.raw_height, tile_spec.raw_width)
            if raw
            else (tile_spec.height, tile_spec.width)
        )
        downsample = tile_spec.downsample
        if annotation_key is not None:
            create_anno_mask = True
            if annotation_name is None:
                raise ValueError(
                    "annotation_name must be provided to create annotation mask."
                )
            if annotation_label is None:
                raise ValueError(
                    "annotation_label must be provided to create annotation mask."
                )

            anno_tb = self._obj.shapes[annotation_key]
            if isinstance(annotation_name, str):
                annotation_name = anno_tb[annotation_name]

            if isinstance(annotation_label, str):
                annotation_label = anno_tb[annotation_label]
            else:
                annotation_label = [annotation_label[n] for n in annotation_name]
            annotation_name = np.asarray(annotation_name)
            annotation_label = np.asarray(annotation_label)

            if mask_dtype is None:
                if annotation_label.max() > 255:
                    mask_dtype = np.uint32
                else:
                    mask_dtype = np.uint8

        points = self._obj[key]
        if sample_n is not None:
            points = points.sample(n=sample_n, random_state=seed)
        elif shuffle:
            points = points.sample(frac=1, random_state=seed)

        for _, row in points.iterrows():
            x = row["x"]
            y = row["y"]
            ix = row["id"]
            tix = row["tissue_id"]

            img = self._obj.reader.get_region(
                x, y, tile_spec.raw_width, tile_spec.raw_height, level=tile_spec.level
            )
            img = cn_func(img)

            if not raw:
                img = cv2.resize(img, (tile_spec.width, tile_spec.height))

            if format == "cyx":
                img = img.transpose(2, 0, 1)

            anno_shapes = None
            anno_mask = None
            if create_anno_mask:
                anno_shapes = []
                bbox = box(x, y, x + tile_spec.raw_width, y + tile_spec.raw_height)
                sel = anno_tb.geometry.intersects(bbox)  # return a boolean mask
                anno_mask = np.zeros(mask_size, dtype=mask_dtype)
                if sel.sum() > 0:
                    sel = sel.values
                    geos = anno_tb.geometry[sel]
                    names = annotation_name[sel]
                    labels = annotation_label[sel]

                    for geo, name, label in zip(geos, names, labels):
                        geo = translate(geo, xoff=-x, yoff=-y)
                        if not raw:
                            geo = scale(
                                geo, 1 / downsample, 1 / downsample, origin=(0, 0)
                            )
                        cnt = np.array(geo.exterior.coords, dtype=np.int32)
                        holes = [
                            np.array(h.coords, dtype=np.int32) for h in geo.interiors
                        ]
                        cv2.fillPoly(anno_mask, [cnt], int(label))  # noqa
                        cv2.fillPoly(anno_mask, holes, 0)  # noqa
                        # Clip the annotation by the tile
                        # May not be valid after clipping
                        output_geo = clip_by_rect(geo, 0, 0, *mask_size)
                        if not output_geo.is_valid:
                            continue
                        elif isinstance(output_geo, MultiPolygon):
                            output_geo = [p for p in output_geo.geoms]
                        else:
                            output_geo = [output_geo]
                        for p in output_geo:
                            anno_shapes.append((p, name, label))

            yield TileImage(
                id=ix,
                x=x,
                y=y,
                tissue_id=tix,
                image=img,
                anno_mask=anno_mask,
                anno_shapes=anno_shapes,
            )



"""Data structures and I/O functions for whole-slide images (WSIs)."""

__version__ = "0.2.0"

import wsidata.dataset as dataset
from ._model import WSIData, TileSpec
from .io import open_wsi, agg_wsi
from ._accessors import (
    register_wsidata_accessor,
    FetchAccessor,
    IterAccessor,
    DatasetAccessor,
)
from .reader import get_reader



from pathlib import Path

import fsspec


def _create_progress(pbar=True, known_length=True):
    from rich.progress import (
        Progress,
        BarColumn,
        TextColumn,
        DownloadColumn,
        TransferSpeedColumn,
        TimeRemainingColumn,
        SpinnerColumn,
    )

    if known_length:
        return Progress(
            TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
            BarColumn(bar_width=20),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
            disable=not pbar,
        )
    else:
        return Progress(
            TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
            SpinnerColumn(),
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            disable=not pbar,
        )


class CacheDownloader:
    """A downloader with simple cache mechanism.

    This class can download from arbitrary URLs and cache the downloaded file.

    The hash of the downloaded file is stored in a hidden file
    in the same directory of the downloaded file.

    Parameters
    ----------
    url : str
        URL to download.
    name : str, optional
        Name of the file, by default None.
    cache_dir : str, optional
        Cache directory, by default None.

    """

    hash_method = "sha256"

    def __init__(self, url, name=None, cache_dir=None):
        self.url = url
        with fsspec.open(self.url, "rb") as fsrc:
            self.total_size = fsrc.size  # Retrieve the total file size
            if name is None:
                if hasattr(fsrc, "name"):
                    name = fsrc.name
        if name is None:
            name = Path(url).name
        if cache_dir is None:
            cache_dir = "."  # Default to current directory
        cache_dir = Path(cache_dir)
        self.name = name
        self.dest = cache_dir / name
        self.hash_path = cache_dir / f".{name}.{self.hash_method}"

    def is_cache(self):
        if self.dest.exists() & self.hash_path.exists():
            with open(self.hash_path, "r") as f:
                last_file_hash = f.read()
            with open(self.dest, "rb") as f:
                current_file_hash = self._hash_file(f)
            return last_file_hash == current_file_hash
        return False

    def _hash_file(self, fileobj):
        import hashlib

        if hasattr(hashlib, "file_digest"):
            digest = hashlib.file_digest(fileobj, self.hash_method)
            hex_digest = digest.hexdigest()
        else:
            h = getattr(hashlib, self.hash_method)()
            b = bytearray(128 * 1024)
            mv = memoryview(b)
            while n := fileobj.readinto(mv):
                h.update(mv[:n])
            hex_digest = h.hexdigest()

        return hex_digest

    def download(self, pbar=True):
        """Download a single file with progress tracking."""
        if self.is_cache():
            return self.dest
        else:
            known_length = self.total_size is not None
            progress = _create_progress(pbar, known_length)
            with progress:
                with fsspec.open(self.url, "rb") as fsrc:
                    task_id = progress.add_task(
                        "Download", filename=self.dest, total=self.total_size
                    )
                    with fsspec.open(f"{self.dest}", "wb") as fdst:
                        progress.start_task(task_id)
                        chunk_size = 1024 * 1024  # 1 MB
                        while chunk := fsrc.read(chunk_size):
                            fdst.write(chunk)
                            progress.advance(task_id, len(chunk))
                progress.refresh()
                # Create a hash file
                with open(self.hash_path, "w") as f:
                    with open(self.dest, "rb") as fdst:
                        f.write(self._hash_file(fdst))
            return self.dest
